﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using symdump.util;

namespace symdump.symfile
{
    public class UnionDef : IEquatable<UnionDef>
    {
        private List<StructMember> _members = new List<StructMember>();
        private string _name;
        public readonly int SYMFileSequenceNumber;

        public List<StructMember> Members
        {
            get
            {
                return _members;
            }
            set
            {
                _members = value;
            }
        }

        public string Name
        {
            get
            {
                return _name;
            }
            set
            {
                _name = value;
            }
        }

        public UnionDef(BinaryReader stream, string name, BenLincoln.DebugSymbols.SequenceIssuer issuer)
        {
            _name = name;
            SYMFileSequenceNumber = issuer.GetNextNumber();
            while (true)
            {
                var typedValue = new TypedValue(stream, issuer);
                if (typedValue.Type == (0x80 | 20))
                {
                    var m = new StructMember(typedValue, stream, false, string.Format("union {0}", this.Name), this.GetHashCode(), issuer);

                    if (m.TypeInfo.ClassType == ClassType.EndOfStruct)
                        break;

                    _members.Add(m);
                }
                else if (typedValue.Type == (0x80 | 22))
                {
                    var m = new StructMember(typedValue, stream, true, string.Format("union {0}", this.Name), this.GetHashCode(), issuer);

                    if (m.TypeInfo.ClassType == ClassType.EndOfStruct)
                        break;

                    _members.Add(m);
                }
                else
                {
                    throw new Exception("Unexpected entry");
                }
            }
        }

        public bool IsFake => new Regex(@"^\.\d+fake$").IsMatch(_name);

        public bool Equals(UnionDef other)
        {
            if (ReferenceEquals(null, other)) return false;
            if (ReferenceEquals(this, other)) return true;
            return _members.SequenceEqual(other._members) && string.Equals(_name, other._name);
        }

        public void Dump(IndentedTextWriter writer)
        {
            writer.WriteLine($"union {_name} {{");
            ++writer.Indent;
            foreach (var m in _members)
                writer.WriteLine(m);
            --writer.Indent;
            writer.WriteLine("};");
        }

        public BenLincoln.DebugSymbols.SymDump.CUnion ToBLDSCUnion()
        {
            BenLincoln.DebugSymbols.SymDump.CUnion result = new BenLincoln.DebugSymbols.SymDump.CUnion();
            result.name = _name;
            result.hashcode = GetHashCode();
            result.members = new List<BenLincoln.DebugSymbols.SymDump.CStructMember>();
            result.sym_file_sequence_number = SYMFileSequenceNumber;
            foreach (StructMember m in _members)
            {
                result.members.Add(m.ToBLDSCStructMember());
            }

            return result;
        }

        public string DumpJSON(string sourceFile, string lineIndent)
        {
            BenLincoln.DebugSymbols.SymDump.CUnion resultDef = ToBLDSCUnion();

            return BenLincoln.DebugSymbols.Utilities.IndentString(resultDef.ToJSON(), lineIndent);
        }

        public override bool Equals(object obj)
        {
            if (ReferenceEquals(null, obj)) return false;
            if (ReferenceEquals(this, obj)) return true;
            if (obj.GetType() != GetType()) return false;
            return Equals((UnionDef) obj);
        }

        public override int GetHashCode()
        {
            unchecked
            {
                return ((_members != null ? _members.GetHashCode() : 0) * 397) ^
                       (_name != null ? _name.GetHashCode() : 0);
            }
        }
    }
}
